//---------------------------------------------------------------------------

#pragma hdrstop

#define _EXPORTING
#include "..\miniscope\device\MiniscopeDevice.h"
#include "..\miniscope\device\MiniscopeDeviceCapabilities.h"
#include <assert.h>
#include "DirectSoundScope.h"
#include "FormSettings.h"
#include "PluginSettings.h"
#include <algorithm>	// needed by Utils::in_group
#include "Utils.h"
#include <string>
#include <fstream> 
#include <json/json.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)

static const struct S_DEVICE_DLL_INTERFACE dll_interface =
	{DLL_INTERFACE_MAJOR_VERSION, DLL_INTERFACE_MINOR_VERSION};

// callback ptrs
CALLBACK_LOG lpLogFn = NULL;
CALLBACK_CONNECT lpConnectFn = NULL;
CALLBACK_DATA lpDataFn = NULL;
CALLBACK_TRIGGER lpTriggerFn = NULL;


void *callbackCookie;	///< used by upper class to distinguish library instances when receiving callbacks

void __stdcall GetMiniscopeInterfaceDescription(struct S_DEVICE_DLL_INTERFACE* interf)
{
	interf->majorVersion = dll_interface.majorVersion;
	interf->minorVersion = dll_interface.minorVersion;
}

void Log(char* txt)
{
	if (lpLogFn)
		lpLogFn(callbackCookie, txt);
}

void Data(short* data, int size)
{
	if (lpDataFn)
		lpDataFn(callbackCookie, data, size);
}

void Connect(int state, char *szMsgText)
{
	if (lpConnectFn)
		lpConnectFn(callbackCookie, state, szMsgText);
}

void Trigger(void)
{
	if (lpTriggerFn)
		lpTriggerFn(callbackCookie);
}

void __stdcall SetCallbacks(void *cookie, CALLBACK_LOG lpLog, CALLBACK_CONNECT lpConnect,
	CALLBACK_DATA lpData, CALLBACK_TRIGGER lpTrigger)
{
	assert(cookie && lpLog && lpConnect && lpData && lpTrigger);
	lpLogFn = lpLog;
	lpConnectFn = lpConnect;
	lpDataFn = lpData;
	lpTriggerFn = lpTrigger;
	callbackCookie = cookie;
	lpLogFn(callbackCookie, "dsoundscope.dll loaded. Yay!");

	scope.callbackLog = Log;
	scope.callbackData = Data;
	scope.callbackConnect = Connect;
	scope.callbackTrigger = Trigger;
}

static enum E_TRIGGER_MODE eMode[] = { TRIGGER_MODE_AUTO, TRIGGER_MODE_NORMAL, TRIGGER_MODE_SINGLE, TRIGGER_MODE_CONTINUOUS };
static enum E_TRIGGER_TYPE eTrigger[] = { TRIGGER_SLOPE_UP, TRIGGER_SLOPE_DOWN };
static enum E_COUPLING_TYPE eCoupling[] = { COUPLING_AC };

/** \sa size of DirectSoundScope::output_buffer when increasing max size
*/
static unsigned int iBuf[] = { 22050, 44100, 88200 };
static float fSens[] = { 3.052e-5 };
const float fBasicSampl = 2.26757e-5;	// 44100
static float fSampl[] = { fBasicSampl*256, fBasicSampl*128, fBasicSampl*64, fBasicSampl*32,
	fBasicSampl*16, fBasicSampl*8, fBasicSampl*4, fBasicSampl*2, fBasicSampl };

enum { CHANNEL_CNT = DirectSoundScope::CHANNEL_CNT };
bool pChannelEnable[CHANNEL_CNT];
unsigned int pSensitivity[CHANNEL_CNT];
enum E_COUPLING_TYPE eCouplingType[CHANNEL_CNT];

/** \brief Calibration data for single sensitivity range
*/
static struct S_CALIBRATION
{
	int offset;
	float gain;
} calibration[sizeof(fSens)/sizeof(fSens[0])];

void __stdcall GetDeviceCapabilities(struct S_SCOPE_CAPABILITIES **caps)
{
	static struct S_SCOPE_CAPABILITIES capabilities =
	{
        CHANNEL_CNT,
		16,		//bits/sample
		sizeof(fSens)/sizeof(fSens[0]), fSens,
		sizeof(fSampl)/sizeof(fSampl[0]), fSampl,
		sizeof(eMode)/sizeof(eMode[0]), eMode,
		sizeof(eTrigger)/sizeof(eTrigger[0]), eTrigger,
		sizeof(eCoupling)/sizeof(eCoupling[0]), eCoupling,
		sizeof(iBuf)/sizeof(iBuf[0]), iBuf
	};
	*caps = &capabilities;
}

void __stdcall ShowSettings(HANDLE parent)
{
	//MessageBox(parent, "Dsoundscope.dll has no additional settings.", "Device DLL", MB_ICONINFORMATION);
	TfrmSettings* frmSettings = new TfrmSettings(NULL);
	std::vector<std::string> devices;
	scope.GetCaptureDevicesList(devices);
	frmSettings->SetDevicesList(devices);
	frmSettings->ShowModal();
	if (scope.GetDevice() != pluginSettings.asCaptureDevice)
	{
		scope.SetDevice(pluginSettings.asCaptureDevice);
		scope.Connect();
	}                   
	delete frmSettings;
}

int __stdcall SetSensitivity(int iChannelId, int iId)
{
	if (iChannelId < 0 || iChannelId >= CHANNEL_CNT)
		return -1;
	if (iId < 0 || iId >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -2;
	return scope.SetSensitivity(iChannelId, iId);
}

int __stdcall SetSamplingPeriod(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(fSampl)/sizeof(fSampl[0])))
		return -1;
	return scope.SetSampling(sizeof(fSampl)/sizeof(fSampl[0]) - iId - 1);
}

int __stdcall SetTriggerMode(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(eMode)/sizeof(eMode[0])))
		return -1;
	return scope.SetTriggerMode(eMode[iId]);
}

int __stdcall ManualTrigger(void)
{
	return scope.ManualTrigger();
}

int __stdcall SetTriggerSource(int iChannelId, int iId)
{
	if (iChannelId < 0 || iChannelId >= CHANNEL_CNT)
		return -1;
	if (iId < 0 || iId >= (int)(sizeof(eTrigger)/sizeof(enum E_TRIGGER_TYPE)))
		return -2;
	return scope.SetTriggerType(iChannelId, eTrigger[iId]);
}

int __stdcall SetTriggerLevel(int level)
{
	return scope.SetTriggerLevel(level);	
}

int __stdcall SetBufferSize(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(iBuf)/sizeof(unsigned int)))
		return -1;
	return scope.SetBufferSize(iBuf[iId]);
}


int __stdcall Connect(void)
{
	return scope.Connect();
}

int __stdcall Disconnect(void)
{
	return scope.Disconnect();
}

int __stdcall Run(bool state)
{
	return scope.Run(state);
}

static bool bSettingsReaded = false;

static int GetDefaultSettings(struct S_SCOPE_SETTINGS* settings)
{
	settings->pChannelEnable = pChannelEnable;
	settings->pSensitivity = pSensitivity;
	settings->eCouplingType = eCouplingType;

	settings->pChannelEnable[0] = true;
	settings->eCouplingType[0] = COUPLING_DC;
	settings->pSensitivity[0] = 0;
	settings->iSamplingPeriod = 0;
	settings->eTriggerMode = TRIGGER_MODE_AUTO;
	settings->eTriggerType = TRIGGER_SLOPE_UP;
	settings->iTriggerLevel = 0;
	settings->iTriggerPosition = 50;
	settings->iTriggerSrcChannel = 0;
	settings->iBufferSize = 4096;

	for (unsigned int i=0; i<sizeof(calibration)/sizeof(calibration[0]); i++)
	{
		calibration[i].offset = 0;
		calibration[i].gain = 1.0;
	}
	bSettingsReaded = true;
	return 0;
}

int __stdcall GetMiniscopeSettings(struct S_SCOPE_SETTINGS* settings)
{
	settings->pChannelEnable = pChannelEnable;
	settings->pSensitivity = pSensitivity;
	settings->eCouplingType = eCouplingType;

	settings->iTriggerSrcChannel = 0;

	std::string path = Utils::GetDllPath();
	path = Utils::ReplaceFileExtension(path, ".cfg");
	if (path == "")
		return GetDefaultSettings(settings);

	Json::Value root;   // will contains the root value after parsing.
	Json::Reader reader;

	std::ifstream ifs(path.c_str());
	std::string strConfig((std::istreambuf_iterator<char>(ifs)), std::istreambuf_iterator<char>());
	ifs.close();

	bool parsingSuccessful = reader.parse( strConfig, root );
	if ( !parsingSuccessful )
		return GetDefaultSettings(settings);

	settings->pSensitivity[0] = root.get("Sensitivity", 0).asInt();
	if (settings->pSensitivity[0] >= sizeof(fSens)/sizeof(fSens[0]))
		settings->pSensitivity[0] = 0;

	settings->iSamplingPeriod = root.get("SamplingPeriod", 0).asInt();
	if (settings->iSamplingPeriod >= sizeof(fSampl)/sizeof(fSampl[0]))
		settings->iSamplingPeriod = 0;

	int mode = root.get("TriggerMode", TRIGGER_MODE_AUTO).asInt();
	if (Utils::in_group(mode, eMode))
		settings->eTriggerMode = (E_TRIGGER_MODE)mode;
	else
		settings->eTriggerMode = TRIGGER_MODE_AUTO;

	int trigger = root.get("TriggerType", TRIGGER_SLOPE_UP).asInt();
	if (Utils::in_group(trigger, eTrigger))
		settings->eTriggerType = (E_TRIGGER_TYPE)trigger;
	else
		settings->eTriggerType = TRIGGER_SLOPE_UP;

	settings->iTriggerLevel = root.get("TriggerLevel", 0).asInt();
	if (settings->iTriggerLevel < -100 || settings->iTriggerLevel > 100)
		settings->iTriggerLevel = 0;

	int coupling = root.get("CouplingType", COUPLING_AC).asInt();
	if (Utils::in_group(coupling, eCoupling))
		settings->eCouplingType[0] = (E_COUPLING_TYPE)coupling;
	else
		settings->eCouplingType[0] = COUPLING_AC;

	settings->iBufferSize = root.get("BufferSize", 4096).asInt();
	if (!Utils::in_group(settings->iBufferSize, iBuf))
		settings->iBufferSize = iBuf[0];

	// read calibration data also
	const Json::Value rootcal = root["Calibration"];	
	for (unsigned int i=0; i<sizeof(calibration)/sizeof(calibration[0]); i++)
	{
		calibration[i].offset = rootcal[i].get("Offset", 0).asInt();
		calibration[i].gain = rootcal[i].get("Gain", 1.0).asDouble();
	}

	// read plugin settings
	const Json::Value rootPluginSettings = root["PluginSettings"];
	pluginSettings.asCaptureDevice = rootPluginSettings.get("CaptureDevice", "").asString().c_str();

	scope.SetDevice(pluginSettings.asCaptureDevice);

	bSettingsReaded = true;
	return 0;
}

int __stdcall SaveMiniscopeSettings(struct S_SCOPE_SETTINGS* settings)
{
    Json::Value root;
	Json::StyledWriter writer;
	root["Sensitivity"] = settings->pSensitivity[0];
	root["SamplingPeriod"] = settings->iSamplingPeriod;
    root["TriggerMode"] = settings->eTriggerMode;
	root["TriggerType"] = settings->eTriggerType;
	root["TriggerLevel"] = settings->iTriggerLevel;
	root["CouplingType"] = settings->eCouplingType[0];
	root["BufferSize"] = settings->iBufferSize;

	// add calibration data also
	const Json::Value rootcal = root["Calibration"];
	for (unsigned int i=0; i<sizeof(calibration)/sizeof(calibration[0]); i++)
	{
		  root["Calibration"][i]["Offset"] = calibration[i].offset;
		  root["Calibration"][i]["Gain"] = calibration[i].gain;
	}

	// store plugin settings
	pluginSettings.asCaptureDevice = scope.GetDevice();
    root["PluginSettings"]["CaptureDevice"] = pluginSettings.asCaptureDevice.c_str();


	std::string outputConfig = writer.write( root );

	std::string path = Utils::GetDllPath();
	path = Utils::ReplaceFileExtension(path, ".cfg");
	if (path == "")
		return -1;

    std::ofstream ofs(path.c_str());
    ofs << outputConfig;
	ofs.close();

	return 0;
}

int __stdcall StoreCalibration(int channel, int range, int offset, float gain)
{
	if (channel < 0 || channel >= CHANNEL_CNT)
		return -1;
	if (range < 0 || range >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -1;
	calibration[range].offset = offset;
	calibration[range].gain = gain;
	return 0;
}

int __stdcall GetCalibration(int channel, int range, int* const offset, float* const gain)
{
	if (channel < 0 || channel >= CHANNEL_CNT)
		return -1;
	if (range < 0 || range >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -2;
	struct S_SCOPE_SETTINGS settings;
	if (bSettingsReaded == false)
	{
		int status = GetMiniscopeSettings(&settings);
		if (status)
			return status;
	}
	*offset = calibration[range].offset;
	*gain = calibration[range].gain;
	return 0;
}

